from google.auth.transport.requests import Request
from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow

import os
import pickle

GMAIL_AUTH_SECRETS_PATH = "~/.config/dmarc"


class GmailAuth:
    def __init__(
        self,
        scopes,
        clientIdFileName=f"{GMAIL_AUTH_SECRETS_PATH}/client_id.json",
        tokenFileName=f"{GMAIL_AUTH_SECRETS_PATH}/token.pickle",
        serviceAccountIdFileName=f"{GMAIL_AUTH_SECRETS_PATH}/service_id.json",
    ):
        self.clientIdFileName = os.path.expanduser(clientIdFileName)
        self.scopes = scopes
        self.tokenFileName = os.path.expanduser(tokenFileName)
        self.serviceAccountIdFileName = os.path.expanduser(serviceAccountIdFileName)

    def authenticateWithClientAccount(self):
        creds = None
        # The file token.pickle stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first
        # time.
        if os.path.exists(self.tokenFileName):
            with open(self.tokenFileName, 'rb') as token:
                creds = pickle.load(token)

        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    self.clientIdFileName, self.scopes
                )
                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open(self.tokenFileName, 'wb') as token:
                pickle.dump(creds, token)

        return build('gmail', 'v1', credentials=creds)

    def authenticateWithServiceAccount(self, fromEmail):
        from google.oauth2 import service_account

        creds = service_account.Credentials.from_service_account_file(
            self.serviceAccountIdFileName, scopes=self.scopes
        )
        return build('gmail', 'v1', credentials=creds.with_subject(fromEmail))



import argparse
import sys
import os

import filetype

from DownloadDMARCFilesFromGmail.GmailAuth import GmailAuth
from DownloadDMARCFilesFromGmail.GmailLabelQuery import GmailLabelQuery


def main():
    parser = argparse.ArgumentParser(
        description='Download DMARC attachments from your Gmail account. '
        'The emails are found based on a label, '
        'and once the download is successful the processed label is applied. '
        'First you need to setup a Google Cloud Console app that allows access to Gmail API. '
        'Visit https://console.cloud.google.com to set it up, '
        'copy the authentication file to the secrets folder, '
        'and rename it to "client_id.json".'
    )
    parser.add_argument(
        '--labelName',
        type=str,
        help='The Gmail label assigned to DMARC emails. Default is "DMARC"',
        default="DMARC",
    )
    parser.add_argument(
        '--processedLabelName',
        type=str,
        help='The Gmail label that will be assigned to processed DMARC emails. Default is "PROCESSED_DMARC"',
        default="PROCESSED_DMARC",
    )
    parser.add_argument(
        '--output',
        type=str,
        help='The location where to store the downloaded DMARC reports. Default is "./out"',
        default="out",
    )

    args = parser.parse_args()

    labelName = args.labelName
    processedLabelName = args.processedLabelName
    downloadPath = args.output
    try:
        service = authenticate()
    except Exception as e:
        print(e)
        sys.exit(
            "ERROR: Authentication failed. Have you enabled access to Gmail API? "
            "Check https://github.com/MozaicWorks/DownloadDMARCAttachmentsFromGmail#how-to-use for instructions"
        )

    labelQuery = GmailLabelQuery(service)
    labelId = labelQuery.labelIdFromName(labelName)
    replaceWithLabelId = labelQuery.labelIdFromName(processedLabelName)

    downloadAttachments(service, labelId, replaceWithLabelId, downloadPath)


def downloadAttachments(service, labelId, replaceWithLabelId, downloadPath):
    messages = messagesWithLabel(service, labelId)
    print("Found {messageCount} messages".format(messageCount=len(messages)))
    createDownloadPath(downloadPath)

    for message in messages:
        try:
            processMessage(
                service, message["id"], labelId, replaceWithLabelId, downloadPath
            )
        except Exception as e:
            print("Error at message {message}".format(message=message))
            print(repr(e))


def processMessage(service, messageId, labelId, replaceWithLabelId, downloadPath):
    downloadAttachment(service, messageId, downloadPath)
    markMessageAsProcessed(service, messageId, labelId, replaceWithLabelId)


def createDownloadPath(downloadPath):
    if not os.path.exists(downloadPath):
        os.makedirs(downloadPath)


def authenticate():
    scopes = ['https://www.googleapis.com/auth/gmail.modify']
    return GmailAuth(scopes).authenticateWithClientAccount()


def markMessageAsProcessed(service, messageId, labelId, replaceWithLabelId):
    modifyMessageRequest = {
        'removeLabelIds': [labelId],
        'addLabelIds': [replaceWithLabelId],
    }
    service.users().messages().modify(
        userId="me", id=messageId, body=modifyMessageRequest
    ).execute()


def labelIdFromName(service, name):
    results = service.users().labels().list(userId="me").execute()
    labels = results.get('labels', [])
    return next(it['id'] for it in labels if it['name'] == name)


def messagesWithLabel(service, labelId):
    MAX_ALLOWED_MESSAGES_COUNT = 500
    result = (
        service.users()
        .messages()
        .list(userId="me", maxResults=MAX_ALLOWED_MESSAGES_COUNT, labelIds=[labelId])
        .execute()
    )
    return result.get("messages", [])


def downloadAttachment(service, messageId, downloadPath):
    import base64

    message = service.users().messages().get(userId="me", id=messageId).execute()
    if "filename" in message["payload"] and "parts" not in message["payload"]:
        fileName = message["payload"]["filename"]
        if not fileName:
            fileName = f"noname-{messageId}"
        print("Processing file {fileName}".format(fileName=fileName))
        attachmentId = message["payload"]["body"]["attachmentId"]
    else:
        part = next(
            it
            for it in message["payload"]["parts"]
            if it["mimeType"] == "application/octet-stream"
            or it["mimeType"] == "application/gzip"
            or it["mimeType"] == "application/zip"
        )
        fileName = part["filename"]
        print("Processing file {fileName}".format(fileName=fileName))
        attachmentId = part["body"]["attachmentId"]

    attachment = (
        service.users()
        .messages()
        .attachments()
        .get(userId="me", messageId=messageId, id=attachmentId)
        .execute()
    )
    data = attachment['data']

    file_data = base64.urlsafe_b64decode(data.encode('UTF-8'))

    filePath = os.path.join(downloadPath, fileName)
    with open(filePath, 'wb') as f:
        f.write(file_data)
    ext = os.path.splitext(filePath)[1]
    if not ext:
        print(f"Determining file extension for {fileName}")
        kind = filetype.guess(filePath)
        fileName = f"{fileName}.{kind.extension}"
        newFilePath = os.path.join(downloadPath, fileName)
        os.rename(filePath, newFilePath)

    print("File {fileName} written".format(fileName=fileName))


if __name__ == "__main__":
    main()



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
Modified by Madoshakalaka@Github (dependency links added)
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path

# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name="DownloadDMARCAttachmentsFromGmail",  # Required
    version="0.2.0",  # Required
    description="Downloads DMARC attachments from Gmail",  # Optional
    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional
    # Denotes that our long_description is in Markdown; valid values are
    # text/plain, text/x-rst, and text/markdown
    #
    # Optional if long_description is written in reStructuredText (rst) but
    # required for plain-text or Markdown; if unspecified, "applications should
    # attempt to render [the long_description] as text/x-rst; charset=UTF-8 and
    # fall back to text/plain if it is not valid rst" (see link below)
    #
    # This field corresponds to the "Description-Content-Type" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-content-type-optional
    long_description_content_type="text/markdown",  # Optional (see note above)
    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url="https://github.com/MozaicWorks/DownloadDMARCAttachmentsFromGmail",  # Optional
    # This should be your name or the name of the organization which owns the
    # project.
    author="Mozaic Works",  # Optional
    # This should be a valid email address corresponding to the author listed
    # above.
    author_email="alex.bolboaca@mozaicworks.com",  # Optional
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Communications :: Email",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # These classifiers are *not* checked by 'pip install'. See instead
        # 'python_requires' below.
        "Programming Language :: Python :: 3.10",
    ],
    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords="DMARC Gmail",  # Optional
    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=["contrib", "docs", "tests"]),  # Required
    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. If you
    # do not support Python 2, you can simplify this to '>=3.5' or similar, see
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires=">=3.10",
    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "cachetools==5.3.0; python_version ~= '3.7'",
        "certifi==2022.12.7; python_version >= '3.6'",
        "charset-normalizer==3.0.1; python_version >= '3.6'",
        "google-api-core==2.11.0; python_version >= '3.7'",
        "google-api-python-client==2.77.0",
        "google-auth==2.16.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4, 3.5'",
        "google-auth-httplib2==0.1.0",
        "google-auth-oauthlib==1.0.0",
        "googleapis-common-protos==1.58.0; python_version >= '3.7'",
        "httplib2==0.21.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "idna==3.4; python_version >= '3.5'",
        "oauthlib==3.2.2; python_version >= '3.6'",
        "protobuf==4.21.12; python_version >= '3.7'",
        "pyasn1==0.4.8",
        "pyasn1-modules==0.2.8",
        "pyparsing==3.0.9; python_version >= '3.1'",
        "requests==2.28.2; python_version >= '3.7' and python_version < '4'",
        "requests-oauthlib==1.3.1; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "rsa==4.9; python_version >= '3.6'",
        "six==1.16.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "uritemplate==4.1.1; python_version >= '3.6'",
        "urllib3==1.26.14; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4, 3.5'",
    ],  # Optional
    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={"dev": []},  # Optional
    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # Sometimes you’ll want to use packages that are properly arranged with
    # setuptools, but are not published to PyPI. In those cases, you can specify
    # a list of one or more dependency_links URLs where the package can
    # be downloaded, along with some additional hints, and setuptools
    # will find and install the package correctly.
    # see https://python-packaging.readthedocs.io/en/latest/dependencies.html#packages-not-on-pypi
    #
    dependency_links=[],
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    # package_data={"sample": ["package_data.dat"]},  # Optional
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[("my_data", ["data/data_file"])],  # Optional
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    entry_points={  # Optional
        "console_scripts": [
            "DownloadDMARCFilesFromGmail=DownloadDMARCFilesFromGmail:main"
        ]
    },
    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={  # Optional
        "Bug Reports": "https://github.com/MozaicWorks/DownloadDMARCAttachmentsFromGmail/issues",
        # "Funding": "https://donate.pypi.org",
        # "Say Thanks!": "http://saythanks.io/to/example",
        "Source": "https://github.com/MozaicWorks/DownloadDMARCAttachmentsFromGmail/",
    },
)



class GmailLabelQuery:
    def __init__(self, service):
        self.service = service
        self.labels = None

    def __fetchLabels__(self):
        results = self.service.users().labels().list(userId="me").execute()
        self.labels = results.get('labels', [])

    def __getAllLabels__(self):
        if not self.labels:
            self.__fetchLabels__()

        return self.labels

    def labelIdFromName(self, name):
        return next(it['id'] for it in self.__getAllLabels__() if it['name'] == name)
